from ptrace.cpu_info import CPU_X86_64, CPU_I386, CPU_PPC64, CPU_PPC32, CPU_AARCH64, CPU_RISCV32, CPU_RISCV64
from ptrace.os_tools import RUNNING_LINUX, RUNNING_FREEBSD
if RUNNING_LINUX:
    if CPU_X86_64:
        from ptrace.syscall.linux.x86_64 import SYSCALL_NAMES, SOCKET_SYSCALL_NAMES
    elif CPU_I386:
        from ptrace.syscall.linux.i386 import SYSCALL_NAMES, SOCKET_SYSCALL_NAMES
    elif CPU_PPC64:
        from ptrace.syscall.linux.powerpc64 import SYSCALL_NAMES, SOCKET_SYSCALL_NAMES
    elif CPU_PPC32:
        from ptrace.syscall.linux.powerpc32 import SYSCALL_NAMES, SOCKET_SYSCALL_NAMES
    elif CPU_AARCH64:
        from ptrace.syscall.linux.aarch64 import SYSCALL_NAMES, SOCKET_SYSCALL_NAMES
    elif CPU_RISCV32:
        from ptrace.syscall.linux.riscv32 import SYSCALL_NAMES, SOCKET_SYSCALL_NAMES
    elif CPU_RISCV64:
        from ptrace.syscall.linux.riscv64 import SYSCALL_NAMES, SOCKET_SYSCALL_NAMES
    else:
        raise NotImplementedError("Unsupported CPU architecture")

elif RUNNING_FREEBSD:
    from ptrace.syscall.freebsd_syscall import SYSCALL_NAMES, SOCKET_SYSCALL_NAMES
else:
    SYSCALL_NAMES = {}
    SOCKET_SYSCALL_NAMES = set()



from ptrace.syscall.posix_constants import SYSCALL_ARG_DICT

SIGSET_SIZE = 64
FD_SETSIZE = 1024

RLIMIT_RESOURCE = {
    0: "RLIMIT_CPU",
    1: "RLIMIT_FSIZE",
    2: "RLIMIT_DATA",
    3: "RLIMIT_STACK",
    4: "RLIMIT_CORE",
    5: "RLIMIT_RSS",
    6: "RLIMIT_NPROC",
    7: "RLIMIT_NOFILE",
    8: "RLIMIT_MEMLOCK",
    9: "RLIMIT_AS",
    10: "RLIMIT_LOCKS",
    11: "RLIMIT_SIGPENDING",
    12: "RLIMIT_MSGQUEUE",
    13: "RLIMIT_NICE",
    14: "RLIMIT_RTPRIO",
    15: "RLIMIT_NLIMITS",
}

SIGPROCMASK_HOW = {0: "SIG_BLOCK", 1: "SIG_UNBLOCK", 2: "SIG_SETMASK"}

SYSCALL_ARG_DICT.update({
    "getrlimit": {"resource": RLIMIT_RESOURCE},
    "setrlimit": {"resource": RLIMIT_RESOURCE},
    "sigprocmask": {"how": SIGPROCMASK_HOW},
    "rt_sigprocmask": {"how": SIGPROCMASK_HOW},
})



from ctypes import (Structure,
                    c_char, c_short, c_int, c_uint, c_long, c_ulong)

time_t = c_long
suseconds_t = c_long
rlim_t = c_long


class timeval(Structure):
    _fields_ = (
        ("tv_sec", time_t),
        ("tv_usec", suseconds_t),
    )


class timespec(Structure):
    _fields_ = (
        ("tv_sec", time_t),
        ("tv_nsec", c_long),
    )


class pollfd(Structure):
    _fields_ = (
        ("fd", c_int),
        ("events", c_short),
        ("revents", c_short),
    )


class rlimit(Structure):
    _fields_ = (
        ("rlim_cur", rlim_t),
        ("rlim_max", rlim_t),
    )


class new_utsname(Structure):
    _fields_ = (
        ("sysname", c_char * 65),
        ("nodename", c_char * 65),
        ("release", c_char * 65),
        ("version", c_char * 65),
        ("machine", c_char * 65),
        ("domainname", c_char * 65),
    )

# Arch depend


class user_desc(Structure):
    _fields_ = (
        ("entry_number", c_uint),
        ("base_addr", c_ulong),
        ("limit", c_uint),
        ("_bits_", c_char),
        # unsigned int  seg_32bit:1;
        # unsigned int  contents:2;
        # unsigned int  read_exec_only:1;
        # unsigned int  limit_in_pages:1;
        # unsigned int  seg_not_present:1;
        # unsigned int  useable:1;
    )



"""
Functions and variables to access to Linux proc directory.

Constant:

   - PAGE_SIZE: size of a memory page
"""
from os import readlink, listdir
from resource import getpagesize
from ptrace.tools import timestampUNIX
from datetime import timedelta

PAGE_SIZE = getpagesize()


class ProcError(Exception):
    """
    Linux proc directory error.
    """
    pass


def openProc(path):
    """
    Open a proc entry in read only mode.
    """
    filename = "/proc/%s" % path
    try:
        return open(filename)
    except IOError as err:
        raise ProcError("Unable to open %r: %s" % (filename, err))


def readProc(path):
    """
    Read the content of a proc entry.
    Eg. readProc("stat") to read /proc/stat.
    """
    with openProc(path) as procfile:
        return procfile.read()


def readProcessProc(pid, key):
    """
    Read the content of a process entry in the proc directory.
    Eg. readProcessProc(pid, "status") to read /proc/pid/status.
    """
    try:
        filename = "/proc/%s/%s" % (pid, key)
        with open(filename) as proc:
            return proc.read()
    except IOError as err:
        raise ProcError("Process %s doesn't exist: %s" % (pid, err))


class ProcessState(object):
    """
    Processus state. Attributes:
    - state (str): process status ('R', 'S', 'T', ...)
    - program (str): program name
    - pid (int): process identifier
    - ppid (int): parent process identifier
    - pgrp (int): process group
    - session (int): session identifier
    - tty_nr (int): tty number
    - tpgid (int)
    - utime (int): user space time (jiffies)
    - stime (int): kernel space time (jiffies)
    - starttime (int): start time
    """
    STATE_NAMES = {
        "R": "running",
        "S": "sleeping",
        "D": "disk",
        "Z": "zombie",
        "T": "traced",
        "W": "paging",
    }

    def __init__(self, stat):
        # pid (program) ... => "pid (program", "..."
        part, stat = stat.rsplit(')', 1)
        self.pid, self.program = part.split('(', 1)
        self.pid = int(self.pid)

        # "state ..." => state, "..."
        stat = stat.split()
        self.state = stat[0]
        stat = [int(item) for item in stat[1:]]

        # Read next numbers
        self.ppid = stat[0]
        self.pgrp = stat[1]
        self.session = stat[2]
        self.tty_nr = stat[3]
        self.tpgid = stat[4]
        self.utime = stat[10]
        self.stime = stat[11]
        self.starttime = stat[18]


def readProcessStat(pid):
    """
    Read the process state ('stat') as a ProcessState object.
    """
    stat = readProcessProc(pid, 'stat')
    return ProcessState(stat)


def readProcessStatm(pid):
    """
    Read the process memory status ('statm') as a list of integers.
    Values are in bytes (and not in pages).
    """
    statm = readProcessProc(pid, 'statm')
    statm = [int(item) * PAGE_SIZE for item in statm.split()]
    return statm


def readProcessProcList(pid, key):
    """
    Read a process entry as a list of strings.
    """
    data = readProcessProc(pid, key)
    if not data:
        # Empty file: empty list
        return []
    data = data.split("\0")
    if not data[-1]:
        del data[-1]
    return data


def readProcessLink(pid, key):
    """
    Read a process link.
    """
    try:
        filename = "/proc/%s/%s" % (pid, key)
        return readlink(filename)
    except OSError as err:
        raise ProcError("Unable to read proc link %r: %s" % (filename, err))


def readProcesses():
    """
    Read all processes identifiers. The function is a generator,
    use it with: ::

       for pid in readProcesses(): ...
    """
    for filename in listdir('/proc'):
        try:
            yield int(filename)
        except ValueError:
            # Filename is not an integer (e.g. "stat" from /proc/stat)
            continue


def readProcessCmdline(pid, escape_stat=True):
    """
    Read the process command line. If escape_stat is True, format program name
    with "[%s]" if the process has no command line, e.g. "[khelper]".
    """
    # Try /proc/42/cmdline
    try:
        cmdline = readProcessProcList(pid, 'cmdline')
        if cmdline:
            return cmdline
    except ProcError:
        pass

    # Try /proc/42/stat
    try:
        stat = readProcessStat(pid)
        program = stat.program
        if escape_stat:
            program = "[%s]" % program
        return [program]
    except ProcError:
        return None


def searchProcessesByName(process_name):
    """
    Find all processes matching the program name pattern.
    Eg. pattern "ssh" will find the program "/usr/bin/ssh".

    This function is a generator yielding the process identifier,
    use it with: ::

       for pid in searchProcessByName(pattern):
          ...
    """
    suffix = '/' + process_name
    for pid in readProcesses():
        cmdline = readProcessCmdline(pid)
        if not cmdline:
            continue
        program = cmdline[0]
        if program == process_name or program.endswith(suffix):
            yield pid


def searchProcessByName(process_name):
    """
    Function similar to searchProcessesByName() but only return the identifier
    of the first matching process. Raise a ProcError if there is no matching
    process.
    """
    for pid in searchProcessesByName(process_name):
        return pid
    raise ProcError("Unable to find process: %r" % process_name)


def getUptime():
    """
    Get the system uptime as a datetime.timedelta object.
    """
    uptime = readProc('uptime')
    uptime = uptime.strip().split()
    uptime = float(uptime[0])
    return timedelta(seconds=uptime)


def getSystemBoot():
    """
    Get the system boot date as a datetime.datetime object.
    """
    if getSystemBoot.value is None:
        stat_file = openProc('stat')
        for line in stat_file:
            if not line.startswith("btime "):
                continue
            seconds = int(line[6:])
            btime = timestampUNIX(seconds, True)
            getSystemBoot.value = btime
            break
        stat_file.close()
        if getSystemBoot.value is None:
            raise ProcError("Unable to read system boot time!")
    return getSystemBoot.value


getSystemBoot.value = None



from os import strerror
from ctypes import addressof, c_int, get_errno, set_errno, sizeof
from ptrace import PtraceError
from ptrace.ctypes_tools import formatAddress
from ptrace.os_tools import RUNNING_LINUX, RUNNING_BSD, RUNNING_OPENBSD
from ptrace.cpu_info import CPU_64BITS, CPU_WORD_SIZE, CPU_POWERPC, CPU_AARCH64, CPU_RISCV

if RUNNING_OPENBSD:
    from ptrace.binding.openbsd_struct import (
        reg as ptrace_registers_t,
        fpreg as user_fpregs_struct)

elif RUNNING_BSD:
    from ptrace.binding.freebsd_struct import (
        reg as ptrace_registers_t)

elif RUNNING_LINUX:
    from ptrace.binding.linux_struct import (
        user_regs_struct as ptrace_registers_t,
        user_fpregs_struct, siginfo, iovec_struct)
    if not CPU_64BITS:
        from ptrace.binding.linux_struct import user_fpxregs_struct
else:
    raise NotImplementedError("Unknown OS!")
REGISTER_NAMES = tuple(name for name, type in ptrace_registers_t._fields_)

HAS_PTRACE_SINGLESTEP = True
HAS_PTRACE_EVENTS = False
HAS_PTRACE_IO = False
HAS_PTRACE_SIGINFO = False
HAS_PTRACE_GETREGS = False
HAS_PTRACE_GETREGSET = False
HAS_PTRACE_SETREGS = False
HAS_PTRACE_SETREGSET = False

# Special flags that are required to wait for cloned processes (threads)
# See wait(2)
THREAD_TRACE_FLAGS = 0x00000000

pid_t = c_int

# PTRACE_xxx constants from /usr/include/sys/ptrace.h
# (Linux 2.6.21 Ubuntu Feisty i386)
PTRACE_TRACEME = 0
PTRACE_PEEKTEXT = 1
PTRACE_PEEKDATA = 2
PTRACE_PEEKUSER = 3
PTRACE_POKETEXT = 4
PTRACE_POKEDATA = 5
PTRACE_POKEUSER = 6
PTRACE_CONT = 7
PTRACE_KILL = 8
if HAS_PTRACE_SINGLESTEP:
    PTRACE_SINGLESTEP = 9

if RUNNING_OPENBSD:
    # OpenBSD 4.2 i386
    PTRACE_ATTACH = 9
    PTRACE_DETACH = 10
    HAS_PTRACE_GETREGS = True
    PTRACE_GETREGS = 33
    PTRACE_SETREGS = 34
    PTRACE_GETFPREGS = 35
    PTRACE_SETFPREGS = 36
    HAS_PTRACE_IO = True
    PTRACE_IO = 11
    HAS_PTRACE_SINGLESTEP = True
    PTRACE_SINGLESTEP = 32  # PT_STEP
    # HAS_PTRACE_EVENTS = True
    # PTRACE_SETOPTIONS = 12 # PT_SET_EVENT_MASK
    # PTRACE_GETEVENTMSG = 14 # PT_GET_PROCESS_STATE
elif RUNNING_BSD:
    # FreeBSD 7.0RC1 i386
    PTRACE_ATTACH = 10
    PTRACE_DETACH = 11
    PTRACE_SYSCALL = 22
    if not CPU_POWERPC:
        HAS_PTRACE_GETREGS = True
        PTRACE_GETREGS = 33
    PTRACE_SETREGS = 34
    HAS_PTRACE_IO = True
    PTRACE_IO = 12
else:
    # Linux
    if not (CPU_AARCH64 or CPU_RISCV):
        HAS_PTRACE_GETREGS = True
        HAS_PTRACE_SETREGS = True
        PTRACE_GETREGS = 12
        PTRACE_SETREGS = 13

    HAS_PTRACE_GETREGSET = True
    HAS_PTRACE_SETREGSET = True
    PTRACE_GETREGSET = 0x4204
    PTRACE_SETREGSET = 0x4205
    NT_PRSTATUS = 1

    PTRACE_ATTACH = 16
    PTRACE_DETACH = 17
    PTRACE_SYSCALL = 24
if RUNNING_LINUX:
    PTRACE_GETFPREGS = 14
    PTRACE_SETFPREGS = 15
    if not CPU_64BITS:
        PTRACE_GETFPXREGS = 18
        PTRACE_SETFPXREGS = 19
    HAS_PTRACE_SIGINFO = True
    PTRACE_GETSIGINFO = 0x4202
    PTRACE_SETSIGINFO = 0x4203

    HAS_PTRACE_EVENTS = True
    PTRACE_SETOPTIONS = 0x4200
    PTRACE_GETEVENTMSG = 0x4201

    # Linux introduces the __WALL flag for wait
    THREAD_TRACE_FLAGS = 0x40000000

PTRACE_O_TRACESYSGOOD = 0x00000001
PTRACE_O_TRACEFORK = 0x00000002
PTRACE_O_TRACEVFORK = 0x00000004
PTRACE_O_TRACECLONE = 0x00000008
PTRACE_O_TRACEEXEC = 0x00000010
PTRACE_O_TRACEVFORKDONE = 0x00000020
PTRACE_O_TRACEEXIT = 0x00000040

# Wait extended result codes for the above trace options
PTRACE_EVENT_FORK = 1
PTRACE_EVENT_VFORK = 2
PTRACE_EVENT_CLONE = 3
PTRACE_EVENT_EXEC = 4
PTRACE_EVENT_VFORK_DONE = 5
PTRACE_EVENT_EXIT = 6

try:
    from cptrace import ptrace as _ptrace
    HAS_CPTRACE = True
except ImportError:
    HAS_CPTRACE = False
    from ctypes import c_long, c_ulong
    from ptrace.ctypes_libc import libc

    # Load ptrace() function from the system C library
    _ptrace = libc.ptrace
    _ptrace.argtypes = (c_ulong, c_ulong, c_ulong, c_ulong)
    _ptrace.restype = c_ulong


def ptrace(command, pid=0, arg1=0, arg2=0, check_errno=False):
    if HAS_CPTRACE:
        try:
            set_errno(0)
            result = _ptrace(command, pid, arg1, arg2, check_errno)
        except ValueError as errobj:
            message = str(errobj)
            errno = get_errno()
            raise PtraceError(message, errno=errno, pid=pid)
    else:
        result = _ptrace(command, pid, arg1, arg2)
        result_signed = c_long(result).value
        if result_signed == -1:
            errno = get_errno()
            # peek operations may returns -1 with errno=0:
            # it's not an error. For other operations, -1
            # is always an error
            if not (check_errno) or errno:
                message = "ptrace(cmd=%s, pid=%s, %r, %r) error #%s: %s" % (
                    command, pid, arg1, arg2,
                    errno, strerror(errno))
                raise PtraceError(message, errno=errno, pid=pid)
    return result


def ptrace_traceme():
    ptrace(PTRACE_TRACEME)


def ptrace_attach(pid):
    ptrace(PTRACE_ATTACH, pid)


def ptrace_detach(pid, signal=0):
    ptrace(PTRACE_DETACH, pid, 0, signal)


def _peek(command, pid, address):
    if address % CPU_WORD_SIZE:
        raise PtraceError(
            "ptrace can't read a word from an unaligned address (%s)!"
            % formatAddress(address), pid=pid)
    return ptrace(command, pid, address, check_errno=True)


def _poke(command, pid, address, word):
    if address % CPU_WORD_SIZE:
        raise PtraceError(
            "ptrace can't write a word to an unaligned address (%s)!"
            % formatAddress(address), pid=pid)
    ptrace(command, pid, address, word)


def ptrace_peektext(pid, address):
    return _peek(PTRACE_PEEKTEXT, pid, address)


def ptrace_peekdata(pid, address):
    return _peek(PTRACE_PEEKDATA, pid, address)


def ptrace_peekuser(pid, address):
    return _peek(PTRACE_PEEKUSER, pid, address)


def ptrace_poketext(pid, address, word):
    _poke(PTRACE_POKETEXT, pid, address, word)


def ptrace_pokedata(pid, address, word):
    _poke(PTRACE_POKEDATA, pid, address, word)


def ptrace_pokeuser(pid, address, word):
    _poke(PTRACE_POKEUSER, pid, address, word)


def ptrace_kill(pid):
    ptrace(PTRACE_KILL, pid)


if HAS_PTRACE_EVENTS:
    def WPTRACEEVENT(status):
        return status >> 16

    def ptrace_setoptions(pid, options):
        ptrace(PTRACE_SETOPTIONS, pid, 0, options)

    def ptrace_geteventmsg(pid):
        new_pid = pid_t()
        ptrace(PTRACE_GETEVENTMSG, pid, 0, addressof(new_pid))
        return new_pid.value

if RUNNING_LINUX:
    def ptrace_syscall(pid, signum=0):
        ptrace(PTRACE_SYSCALL, pid, 0, signum)

    def ptrace_cont(pid, signum=0):
        ptrace(PTRACE_CONT, pid, 0, signum)

    def ptrace_getsiginfo(pid):
        info = siginfo()
        ptrace(PTRACE_GETSIGINFO, pid, 0, addressof(info))
        return info

    def ptrace_setsiginfo(pid, info):
        ptrace(PTRACE_SETSIGINFO, pid, 0, addressof(info))

    def ptrace_getfpregs(pid):
        fpregs = user_fpregs_struct()
        ptrace(PTRACE_GETFPREGS, pid, 0, addressof(fpregs))
        return fpregs

    def ptrace_setfpregs(pid, fpregs):
        ptrace(PTRACE_SETFPREGS, pid, 0, addressof(fpregs))

    if not CPU_64BITS:
        def ptrace_getfpxregs(pid):
            fpxregs = user_fpxregs_struct()
            ptrace(PTRACE_GETFPXREGS, pid, 0, addressof(fpxregs))
            return fpxregs

        def ptrace_setfpxregs(pid, fpxregs):
            ptrace(PTRACE_SETFPXREGS, pid, 0, addressof(fpxregs))

    if HAS_PTRACE_GETREGS:
        def ptrace_getregs(pid):
            regs = ptrace_registers_t()
            ptrace(PTRACE_GETREGS, pid, 0, addressof(regs))
            return regs

    elif HAS_PTRACE_GETREGSET:
        def ptrace_getregs(pid):
            regs = ptrace_registers_t()
            iov = iovec_struct()
            setattr(iov, "buf", addressof(regs))
            setattr(iov, "len", sizeof(regs))
            ptrace(PTRACE_GETREGSET, pid, NT_PRSTATUS, addressof(iov))
            return regs

    if HAS_PTRACE_SETREGS:
        def ptrace_setregs(pid, regs):
            ptrace(PTRACE_SETREGS, pid, 0, addressof(regs))

    elif HAS_PTRACE_SETREGSET:
        def ptrace_setregs(pid, regs):
            iov = iovec_struct()
            setattr(iov, "buf", addressof(regs))
            setattr(iov, "len", sizeof(regs))
            ptrace(PTRACE_SETREGSET, pid, NT_PRSTATUS, addressof(iov))

    if HAS_PTRACE_SINGLESTEP:
        def ptrace_singlestep(pid):
            ptrace(PTRACE_SINGLESTEP, pid)

else:
    def ptrace_syscall(pid, signum=0):
        ptrace(PTRACE_SYSCALL, pid, 1, signum)

    def ptrace_cont(pid, signum=0):
        ptrace(PTRACE_CONT, pid, 1, signum)

    if HAS_PTRACE_GETREGS:
        def ptrace_getregs(pid):
            regs = ptrace_registers_t()
            ptrace(PTRACE_GETREGS, pid, addressof(regs))
            return regs

    def ptrace_setregs(pid, regs):
        ptrace(PTRACE_SETREGS, pid, addressof(regs))

    if HAS_PTRACE_SINGLESTEP:
        def ptrace_singlestep(pid):
            ptrace(PTRACE_SINGLESTEP, pid, 1)

if HAS_PTRACE_IO:
    def ptrace_io(pid, io_desc):
        ptrace(PTRACE_IO, pid, addressof(io_desc))



# FreeBSD kernel syscall list from FreeBSD 7.0RC1 for i386
#
# List extracted from:
#    /usr/include/sys/syscall.h
SYSCALL_NAMES = {
    0: "syscall",
    1: "exit",
    2: "fork",
    3: "read",
    4: "write",
    5: "open",
    6: "close",
    7: "wait4",
    #   8: old creat
    9: "link",
    10: "unlink",
    #   11: obsolete execv
    12: "chdir",
    13: "fchdir",
    14: "mknod",
    15: "chmod",
    16: "chown",
    17: "break",
    18: "freebsd4_getfsstat",
    #   19: old lseek
    20: "getpid",
    21: "mount",
    22: "unmount",
    23: "setuid",
    24: "getuid",
    25: "geteuid",
    26: "ptrace",
    27: "recvmsg",
    28: "sendmsg",
    29: "recvfrom",
    30: "accept",
    31: "getpeername",
    32: "getsockname",
    33: "access",
    34: "chflags",
    35: "fchflags",
    36: "sync",
    37: "kill",
    #   38: old stat
    39: "getppid",
    #   40: old lstat
    41: "dup",
    42: "pipe",
    43: "getegid",
    44: "profil",
    45: "ktrace",
    #   46: old sigaction
    47: "getgid",
    #   48: old sigprocmask
    49: "getlogin",
    50: "setlogin",
    51: "acct",
    #   52: old sigpending
    53: "sigaltstack",
    54: "ioctl",
    55: "reboot",
    56: "revoke",
    57: "symlink",
    58: "readlink",
    59: "execve",
    60: "umask",
    61: "chroot",
    #   62: old fstat
    #   63: old getkerninfo
    #   64: old getpagesize
    65: "msync",
    66: "vfork",
    #   67: obsolete vread
    #   68: obsolete vwrite
    69: "sbrk",
    70: "sstk",
    #   71: old mmap
    72: "vadvise",
    73: "munmap",
    74: "mprotect",
    75: "madvise",
    #   76: obsolete vhangup
    #   77: obsolete vlimit
    78: "mincore",
    79: "getgroups",
    80: "setgroups",
    81: "getpgrp",
    82: "setpgid",
    83: "setitimer",
    #   84: old wait
    85: "swapon",
    86: "getitimer",
    #   87: old gethostname
    #   88: old sethostname
    89: "getdtablesize",
    90: "dup2",
    92: "fcntl",
    93: "select",
    95: "fsync",
    96: "setpriority",
    97: "socket",
    98: "connect",
    #   99: old accept
    100: "getpriority",
    #   101: old send
    #   102: old recv
    #   103: old sigreturn
    104: "bind",
    105: "setsockopt",
    106: "listen",
    #   107: obsolete vtimes
    #   108: old sigvec
    #   109: old sigblock
    #   110: old sigsetmask
    #   111: old sigsuspend
    #   112: old sigstack
    #   113: old recvmsg
    #   114: old sendmsg
    #   115: obsolete vtrace
    116: "gettimeofday",
    117: "getrusage",
    118: "getsockopt",
    120: "readv",
    121: "writev",
    122: "settimeofday",
    123: "fchown",
    124: "fchmod",
    #   125: old recvfrom
    126: "setreuid",
    127: "setregid",
    128: "rename",
    #   129: old truncate
    #   130: old ftruncate
    131: "flock",
    132: "mkfifo",
    133: "sendto",
    134: "shutdown",
    135: "socketpair",
    136: "mkdir",
    137: "rmdir",
    138: "utimes",
    #   139: obsolete 4.2 sigreturn
    140: "adjtime",
    #   141: old getpeername
    #   142: old gethostid
    #   143: old sethostid
    #   144: old getrlimit
    #   145: old setrlimit
    #   146: old killpg
    147: "setsid",
    148: "quotactl",
    #   149: old quota
    #   150: old getsockname
    155: "nfssvc",
    #   156: old getdirentries
    157: "freebsd4_statfs",
    158: "freebsd4_fstatfs",
    160: "lgetfh",
    161: "getfh",
    162: "getdomainname",
    163: "setdomainname",
    164: "uname",
    165: "sysarch",
    166: "rtprio",
    169: "semsys",
    170: "msgsys",
    171: "shmsys",
    173: "freebsd6_pread",
    174: "freebsd6_pwrite",
    176: "ntp_adjtime",
    181: "setgid",
    182: "setegid",
    183: "seteuid",
    188: "stat",
    189: "fstat",
    190: "lstat",
    191: "pathconf",
    192: "fpathconf",
    194: "getrlimit",
    195: "setrlimit",
    196: "getdirentries",
    197: "freebsd6_mmap",
    198: "__syscall",
    199: "freebsd6_lseek",
    200: "freebsd6_truncate",
    201: "freebsd6_ftruncate",
    202: "__sysctl",
    203: "mlock",
    204: "munlock",
    205: "undelete",
    206: "futimes",
    207: "getpgid",
    209: "poll",
    220: "__semctl",
    221: "semget",
    222: "semop",
    224: "msgctl",
    225: "msgget",
    226: "msgsnd",
    227: "msgrcv",
    228: "shmat",
    229: "shmctl",
    230: "shmdt",
    231: "shmget",
    232: "clock_gettime",
    233: "clock_settime",
    234: "clock_getres",
    235: "ktimer_create",
    236: "ktimer_delete",
    237: "ktimer_settime",
    238: "ktimer_gettime",
    239: "ktimer_getoverrun",
    240: "nanosleep",
    248: "ntp_gettime",
    250: "minherit",
    251: "rfork",
    252: "openbsd_poll",
    253: "issetugid",
    254: "lchown",
    255: "aio_read",
    256: "aio_write",
    257: "lio_listio",
    272: "getdents",
    274: "lchmod",
    275: "netbsd_lchown",
    276: "lutimes",
    277: "netbsd_msync",
    278: "nstat",
    279: "nfstat",
    280: "nlstat",
    289: "preadv",
    290: "pwritev",
    297: "freebsd4_fhstatfs",
    298: "fhopen",
    299: "fhstat",
    300: "modnext",
    301: "modstat",
    302: "modfnext",
    303: "modfind",
    304: "kldload",
    305: "kldunload",
    306: "kldfind",
    307: "kldnext",
    308: "kldstat",
    309: "kldfirstmod",
    310: "getsid",
    311: "setresuid",
    312: "setresgid",
    #   313: obsolete signanosleep
    314: "aio_return",
    315: "aio_suspend",
    316: "aio_cancel",
    317: "aio_error",
    318: "oaio_read",
    319: "oaio_write",
    320: "olio_listio",
    321: "yield",
    #   322: obsolete thr_sleep
    #   323: obsolete thr_wakeup
    324: "mlockall",
    325: "munlockall",
    326: "__getcwd",
    327: "sched_setparam",
    328: "sched_getparam",
    329: "sched_setscheduler",
    330: "sched_getscheduler",
    331: "sched_yield",
    332: "sched_get_priority_max",
    333: "sched_get_priority_min",
    334: "sched_rr_get_interval",
    335: "utrace",
    336: "freebsd4_sendfile",
    337: "kldsym",
    338: "jail",
    340: "sigprocmask",
    341: "sigsuspend",
    342: "freebsd4_sigaction",
    343: "sigpending",
    344: "freebsd4_sigreturn",
    345: "sigtimedwait",
    346: "sigwaitinfo",
    347: "__acl_get_file",
    348: "__acl_set_file",
    349: "__acl_get_fd",
    350: "__acl_set_fd",
    351: "__acl_delete_file",
    352: "__acl_delete_fd",
    353: "__acl_aclcheck_file",
    354: "__acl_aclcheck_fd",
    355: "extattrctl",
    356: "extattr_set_file",
    357: "extattr_get_file",
    358: "extattr_delete_file",
    359: "aio_waitcomplete",
    360: "getresuid",
    361: "getresgid",
    362: "kqueue",
    363: "kevent",
    371: "extattr_set_fd",
    372: "extattr_get_fd",
    373: "extattr_delete_fd",
    374: "__setugid",
    375: "nfsclnt",
    376: "eaccess",
    378: "nmount",
    379: "kse_exit",
    380: "kse_wakeup",
    381: "kse_create",
    382: "kse_thr_interrupt",
    383: "kse_release",
    384: "__mac_get_proc",
    385: "__mac_set_proc",
    386: "__mac_get_fd",
    387: "__mac_get_file",
    388: "__mac_set_fd",
    389: "__mac_set_file",
    390: "kenv",
    391: "lchflags",
    392: "uuidgen",
    393: "sendfile",
    394: "mac_syscall",
    395: "getfsstat",
    396: "statfs",
    397: "fstatfs",
    398: "fhstatfs",
    400: "ksem_close",
    401: "ksem_post",
    402: "ksem_wait",
    403: "ksem_trywait",
    404: "ksem_init",
    405: "ksem_open",
    406: "ksem_unlink",
    407: "ksem_getvalue",
    408: "ksem_destroy",
    409: "__mac_get_pid",
    410: "__mac_get_link",
    411: "__mac_set_link",
    412: "extattr_set_link",
    413: "extattr_get_link",
    414: "extattr_delete_link",
    415: "__mac_execve",
    416: "sigaction",
    417: "sigreturn",
    421: "getcontext",
    422: "setcontext",
    423: "swapcontext",
    424: "swapoff",
    425: "__acl_get_link",
    426: "__acl_set_link",
    427: "__acl_delete_link",
    428: "__acl_aclcheck_link",
    429: "sigwait",
    430: "thr_create",
    431: "thr_exit",
    432: "thr_self",
    433: "thr_kill",
    434: "_umtx_lock",
    435: "_umtx_unlock",
    436: "jail_attach",
    437: "extattr_list_fd",
    438: "extattr_list_file",
    439: "extattr_list_link",
    440: "kse_switchin",
    441: "ksem_timedwait",
    442: "thr_suspend",
    443: "thr_wake",
    444: "kldunloadf",
    445: "audit",
    446: "auditon",
    447: "getauid",
    448: "setauid",
    449: "getaudit",
    450: "setaudit",
    451: "getaudit_addr",
    452: "setaudit_addr",
    453: "auditctl",
    454: "_umtx_op",
    455: "thr_new",
    456: "sigqueue",
    457: "kmq_open",
    458: "kmq_setattr",
    459: "kmq_timedreceive",
    460: "kmq_timedsend",
    461: "kmq_notify",
    462: "kmq_unlink",
    463: "abort2",
    464: "thr_set_name",
    465: "aio_fsync",
    466: "rtprio_thread",
    471: "sctp_peeloff",
    472: "sctp_generic_sendmsg",
    473: "sctp_generic_sendmsg_iov",
    474: "sctp_generic_recvmsg",
    475: "pread",
    476: "pwrite",
    477: "mmap",
    478: "lseek",
    479: "truncate",
    480: "ftruncate",
    481: "thr_kill2",
}

SOCKET_SYSCALL_NAMES = set((
    "socket", "socketpair", "connect",
    "sendto", "recvfrom", "sendmsg", "recvmsg",
    "bind", "listen", "accept",
    "getpeername", "getsockname", "getsockopt", "setsockopt",
    "shutdown",
))



import socket

SOCKETCALL = {
    1: "socket",
    2: "bind",
    3: "connect",
    4: "listen",
    5: "accept",
    6: "getsockname",
    7: "getpeername",
    8: "socketpair",
    9: "send",
    10: "recv",
    11: "sendto",
    12: "recvfrom",
    13: "shutdown",
    14: "setsockopt",
    15: "getsockopt",
    16: "sendmsg",
    17: "recvmsg",
}

SOCKET_FAMILY = {
    0: "AF_UNSPEC",
    1: "AF_FILE",
    2: "AF_INET",
    3: "AF_AX25",
    4: "AF_IPX",
    5: "AF_APPLETALK",
    6: "AF_NETROM",
    7: "AF_BRIDGE",
    8: "AF_ATMPVC",
    9: "AF_X25",
    10: "AF_INET6",
    11: "AF_ROSE",
    12: "AF_DECnet",
    13: "AF_NETBEUI",
    14: "AF_SECURITY",
    15: "AF_KEY",
    16: "AF_NETLINK",
    17: "AF_PACKET",
    18: "AF_ASH",
    19: "AF_ECONET",
    20: "AF_ATMSVC",
    22: "AF_SNA",
    23: "AF_IRDA",
    24: "AF_PPPOX",
    25: "AF_WANPIPE",
    31: "AF_BLUETOOTH",
}

SOCKET_TYPE = {
    1: "SOCK_STREAM",
    2: "SOCK_DGRAM",
    3: "SOCK_RAW",
    4: "SOCK_RDM",
    5: "SOCK_SEQPACKET",
    10: "SOCK_PACKET",
}


def formatSocketType(argument):
    value = argument.value
    text = []
    if hasattr(socket, 'SOCK_CLOEXEC'):
        cloexec = value & socket.SOCK_CLOEXEC
        value &= ~socket.SOCK_CLOEXEC
    else:
        cloexec = False
    text = SOCKET_TYPE.get(value, str(value))
    if cloexec:
        text += '|SOCK_CLOEXEC'
    return text


SOCKET_PROTOCOL = {
    1: "IPPROTO_ICMP",
    58: "IPPROTO_ICMPV6",
}

SETSOCKOPT_LEVEL = {
    0: "SOL_IP",
    1: "SOL_SOCKET",
}

SETSOCKOPT_OPTNAME = {
    # level 0 (SOL_IP)
    1: "IP_TOS",
    # level 1 (SOL_SOCKET)
    2: "SO_REUSEADDR",
    9: "SO_KEEPALIVE",
    20: "SO_RCVTIMEO",
    21: "SO_SNDTIMEO",
}



"""
Mockup classes used in unit tests.
"""


class FakeProcess(object):

    def __init__(self):
        self.regs = {}

    def setreg(self, name, value):
        self.regs[name] = value

    def getreg(self, name):
        return self.regs[name]



from struct import pack, unpack
from ptrace.cpu_info import CPU_64BITS
from ctypes import cast, POINTER


def int2uint64(value):
    """
    Convert a signed 64 bits integer into an unsigned 64 bits integer.

    >>> print(int2uint64(1))
    1
    >>> print(int2uint64(2**64 + 1))  # ignore bits larger than 64 bits
    1
    >>> print(int2uint64(-1))
    18446744073709551615
    >>> print(int2uint64(-2))
    18446744073709551614
    """
    return (value & 0xffffffffffffffff)


def uint2int64(value):
    """
    Convert an unsigned 64 bits integer into a signed 64 bits integer.

    >>> print(uint2int64(1))
    1
    >>> print(uint2int64(2**64 + 1))  # ignore bits larger than 64 bits
    1
    >>> print(uint2int64(18446744073709551615))
    -1
    >>> print(uint2int64(18446744073709551614))
    -2
    """
    value = value & 0xffffffffffffffff
    if value & 0x8000000000000000:
        return value - 0x10000000000000000
    else:
        return value


def truncateWord32(value):
    """
    Truncate an unsigned integer to 32 bits.
    """
    return value & 0xFFFFFFFF


def truncateWord64(value):
    """
    Truncate an unsigned integer to 64 bits.
    """
    return value & 0xFFFFFFFFFFFFFFFF


def formatUintHex16(value):
    """
    Format an 16 bits unsigned integer.
    """
    return u"0x%04x" % value


def formatUintHex32(value):
    """
    Format an 32 bits unsigned integer.
    """
    return u"0x%08x" % value


def formatUintHex64(value):
    """
    Format an 64 bits unsigned integer.
    """
    return u"0x%016x" % value


def int2uint32(value):
    """
    Convert a signed 32 bits integer into an unsigned 32 bits integer.

    >>> print(int2uint32(1))
    1
    >>> print(int2uint32(2**32 + 1))  # ignore bits larger than 32 bits
    1
    >>> print(int2uint32(-1))
    4294967295
    """
    return value & 0xffffffff


def uint2int32(value):
    """
    Convert an unsigned 32 bits integer into a signed 32 bits integer.

    >>> print(uint2int32(1))
    1
    >>> print(uint2int32(2**32 + 1))  # ignore bits larger than 32 bits
    1
    >>> print(uint2int32(4294967295))
    -1
    >>> print(uint2int32(4294967294))
    -2
    >>> print(uint2int32(18446744073709551615))
    -1
    """
    value = value & 0xffffffff
    if value & 0x80000000:
        v = value - 0x100000000
    else:
        v = value
    return v


uint2int = uint2int32
int2uint = int2uint32
if CPU_64BITS:
    ulong2long = uint2int64
    long2ulong = int2uint64
    formatWordHex = formatUintHex64
    truncateWord = truncateWord64
else:
    ulong2long = uint2int32
    long2ulong = int2uint32
    formatWordHex = formatUintHex32
    truncateWord = truncateWord32


def formatAddress(address):
    """
    Format an address to hexadecimal.
    Return "NULL" for zero.
    """
    if address:
        return formatWordHex(address)
    else:
        return u"NULL"


def formatAddressRange(start, end):
    """
    Format an address range, e.g. "0x80004000-0x8000ffff".
    """
    return u"%s-%s" % (formatWordHex(start), formatWordHex(end))


def ntoh_ushort(value):
    """
    Convert an unsigned short integer from network endian to host endian.
    """
    return unpack("<H", pack(">H", value))[0]


def ntoh_uint(value):
    """
    Convert an unsigned integer from network endian to host endian.
    """
    return unpack("<I", pack(">I", value))[0]


def word2bytes(word):
    """
    Convert an unsigned integer (a CPU word) to a bytes string.
    """
    return pack("L", word)


def bytes2word(bytes):
    """
    Convert a bytes string to an unsigned integer (a CPU word).
    """
    return unpack("L", bytes)[0]


def bytes2type(bytes, type):
    """
    Cast a bytes string to an object of the specified type.
    """
    return cast(bytes, POINTER(type))[0]


def bytes2array(bytes, basetype, size):
    """
    Cast a bytes string to an array of objects of the specified type
    and size.
    """
    return bytes2type(bytes, basetype * size)



from ptrace.error import PtraceError


class ProcessError(PtraceError):

    def __init__(self, process, message):
        PtraceError.__init__(self, message, pid=process.pid)
        self.process = process



from ptrace.binding.func import (   # noqa
    HAS_PTRACE_SINGLESTEP, HAS_PTRACE_EVENTS,
    HAS_PTRACE_IO, HAS_PTRACE_SIGINFO, HAS_PTRACE_GETREGS,
    HAS_PTRACE_GETREGSET, REGISTER_NAMES,
    ptrace_attach, ptrace_traceme,
    ptrace_detach, ptrace_kill,
    ptrace_cont, ptrace_syscall,
    ptrace_setregs,
    ptrace_peektext, ptrace_poketext,
    ptrace_peekuser,
    ptrace_registers_t)
if HAS_PTRACE_EVENTS:
    from ptrace.binding.func import (WPTRACEEVENT,   # noqa
                                     PTRACE_EVENT_FORK, PTRACE_EVENT_VFORK, PTRACE_EVENT_CLONE,
                                     PTRACE_EVENT_EXEC,
                                     ptrace_setoptions, ptrace_geteventmsg)
if HAS_PTRACE_SINGLESTEP:
    from ptrace.binding.func import ptrace_singlestep   # noqa
if HAS_PTRACE_SIGINFO:
    from ptrace.binding.func import ptrace_getsiginfo   # noqa
if HAS_PTRACE_IO:
    from ptrace.binding.func import ptrace_io   # noqa
    from ptrace.binding.freebsd_struct import (   # noqa
        ptrace_io_desc,
        PIOD_READ_D, PIOD_WRITE_D,
        PIOD_READ_I, PIOD_WRITE_I)
if HAS_PTRACE_GETREGS or HAS_PTRACE_GETREGSET:
    from ptrace.binding.func import ptrace_getregs   # noqa



from ctypes import (Structure, Union, sizeof,
                    c_char, c_ushort, c_int, c_uint, c_ulong, c_void_p,
                    c_uint16, c_uint32, c_uint64, c_size_t)
from ptrace.cpu_info import CPU_64BITS, CPU_PPC32, CPU_PPC64, CPU_ARM32, CPU_AARCH64, CPU_RISCV

pid_t = c_int
uid_t = c_ushort
clock_t = c_uint

# From /usr/include/asm-i386/user.h
# Also more reliably in the kernel sources:
# arch/$ARCH/include/uapi/asm/ptrace.h


class register_structure(Structure):
    def __str__(self):
        regs = {}
        for reg in self.__class__._fields_:
            regs.update({reg[0]: getattr(self, reg[0])})
        return str(regs)


class user_regs_struct(register_structure):
    if CPU_PPC32:
        _fields_ = (
            ("gpr0", c_ulong),
            ("gpr1", c_ulong),
            ("gpr2", c_ulong),
            ("gpr3", c_ulong),
            ("gpr4", c_ulong),
            ("gpr5", c_ulong),
            ("gpr6", c_ulong),
            ("gpr7", c_ulong),
            ("gpr8", c_ulong),
            ("gpr9", c_ulong),
            ("gpr10", c_ulong),
            ("gpr11", c_ulong),
            ("gpr12", c_ulong),
            ("gpr13", c_ulong),
            ("gpr14", c_ulong),
            ("gpr15", c_ulong),
            ("gpr16", c_ulong),
            ("gpr17", c_ulong),
            ("gpr18", c_ulong),
            ("gpr19", c_ulong),
            ("gpr20", c_ulong),
            ("gpr21", c_ulong),
            ("gpr22", c_ulong),
            ("gpr23", c_ulong),
            ("gpr24", c_ulong),
            ("gpr25", c_ulong),
            ("gpr26", c_ulong),
            ("gpr27", c_ulong),
            ("gpr28", c_ulong),
            ("gpr29", c_ulong),
            ("gpr30", c_ulong),
            ("gpr31", c_ulong),
            ("nip", c_ulong),
            ("msr", c_ulong),
            ("orig_gpr3", c_ulong),
            ("ctr", c_ulong),
            ("link", c_ulong),
            ("xer", c_ulong),
            ("ccr", c_ulong),
            ("mq", c_ulong),  # FIXME: ppc64 => softe
            ("trap", c_ulong),
            ("dar", c_ulong),
            ("dsisr", c_ulong),
            ("result", c_ulong),
        )
    elif CPU_PPC64:
        _fields_ = (
            ("gpr0", c_ulong),
            ("gpr1", c_ulong),
            ("gpr2", c_ulong),
            ("gpr3", c_ulong),
            ("gpr4", c_ulong),
            ("gpr5", c_ulong),
            ("gpr6", c_ulong),
            ("gpr7", c_ulong),
            ("gpr8", c_ulong),
            ("gpr9", c_ulong),
            ("gpr10", c_ulong),
            ("gpr11", c_ulong),
            ("gpr12", c_ulong),
            ("gpr13", c_ulong),
            ("gpr14", c_ulong),
            ("gpr15", c_ulong),
            ("gpr16", c_ulong),
            ("gpr17", c_ulong),
            ("gpr18", c_ulong),
            ("gpr19", c_ulong),
            ("gpr20", c_ulong),
            ("gpr21", c_ulong),
            ("gpr22", c_ulong),
            ("gpr23", c_ulong),
            ("gpr24", c_ulong),
            ("gpr25", c_ulong),
            ("gpr26", c_ulong),
            ("gpr27", c_ulong),
            ("gpr28", c_ulong),
            ("gpr29", c_ulong),
            ("gpr30", c_ulong),
            ("gpr31", c_ulong),
            ("nip", c_ulong),
            ("msr", c_ulong),
            ("orig_gpr3", c_ulong),
            ("ctr", c_ulong),
            ("link", c_ulong),
            ("xer", c_ulong),
            ("ccr", c_ulong),
            ("softe", c_ulong),
            ("trap", c_ulong),
            ("dar", c_ulong),
            ("dsisr", c_ulong),
            ("result", c_ulong),
        )
    elif CPU_ARM32:
        _fields_ = tuple(("r%i" % reg, c_ulong) for reg in range(18))
    elif CPU_AARCH64:
        _fields_ = tuple([*[("r%i" % reg, c_ulong) for reg in range(31)],
                         ('sp', c_ulong),
                         ('pc', c_ulong),
                         ('pstate', c_ulong)]
                         )
    elif CPU_RISCV:
        _fields_ = (
            ("pc", c_ulong),
            ("ra", c_ulong),
            ("sp", c_ulong),
            ("gp", c_ulong),
            ("tp", c_ulong),
            ("t0", c_ulong),
            ("t1", c_ulong),
            ("t2", c_ulong),
            ("s0", c_ulong),
            ("s1", c_ulong),
            ("a0", c_ulong),
            ("a1", c_ulong),
            ("a2", c_ulong),
            ("a3", c_ulong),
            ("a4", c_ulong),
            ("a5", c_ulong),
            ("a6", c_ulong),
            ("a7", c_ulong),
            ("s2", c_ulong),
            ("s3", c_ulong),
            ("s4", c_ulong),
            ("s5", c_ulong),
            ("s6", c_ulong),
            ("s7", c_ulong),
            ("s8", c_ulong),
            ("s9", c_ulong),
            ("s10", c_ulong),
            ("s11", c_ulong),
            ("t3", c_ulong),
            ("t4", c_ulong),
            ("t5", c_ulong),
            ("t6", c_ulong),
        )
    elif CPU_64BITS:
        _fields_ = (
            ("r15", c_ulong),
            ("r14", c_ulong),
            ("r13", c_ulong),
            ("r12", c_ulong),
            ("rbp", c_ulong),
            ("rbx", c_ulong),
            ("r11", c_ulong),
            ("r10", c_ulong),
            ("r9", c_ulong),
            ("r8", c_ulong),
            ("rax", c_ulong),
            ("rcx", c_ulong),
            ("rdx", c_ulong),
            ("rsi", c_ulong),
            ("rdi", c_ulong),
            ("orig_rax", c_ulong),
            ("rip", c_ulong),
            ("cs", c_ulong),
            ("eflags", c_ulong),
            ("rsp", c_ulong),
            ("ss", c_ulong),
            ("fs_base", c_ulong),
            ("gs_base", c_ulong),
            ("ds", c_ulong),
            ("es", c_ulong),
            ("fs", c_ulong),
            ("gs", c_ulong)
        )
    else:
        _fields_ = (
            ("ebx", c_ulong),
            ("ecx", c_ulong),
            ("edx", c_ulong),
            ("esi", c_ulong),
            ("edi", c_ulong),
            ("ebp", c_ulong),
            ("eax", c_ulong),
            ("ds", c_ushort),
            ("__ds", c_ushort),
            ("es", c_ushort),
            ("__es", c_ushort),
            ("fs", c_ushort),
            ("__fs", c_ushort),
            ("gs", c_ushort),
            ("__gs", c_ushort),
            ("orig_eax", c_ulong),
            ("eip", c_ulong),
            ("cs", c_ushort),
            ("__cs", c_ushort),
            ("eflags", c_ulong),
            ("esp", c_ulong),
            ("ss", c_ushort),
            ("__ss", c_ushort),
        )


class user_fpregs_struct(register_structure):
    if CPU_64BITS:
        _fields_ = (
            ("cwd", c_uint16),
            ("swd", c_uint16),
            ("ftw", c_uint16),
            ("fop", c_uint16),
            ("rip", c_uint64),
            ("rdp", c_uint64),
            ("mxcsr", c_uint32),
            ("mxcr_mask", c_uint32),
            ("st_space", c_uint32 * 32),
            ("xmm_space", c_uint32 * 64),
            ("padding", c_uint32 * 24)
        )
    else:
        _fields_ = (
            ("cwd", c_ulong),
            ("swd", c_ulong),
            ("twd", c_ulong),
            ("fip", c_ulong),
            ("fcs", c_ulong),
            ("foo", c_ulong),
            ("fos", c_ulong),
            ("st_space", c_ulong * 20)
        )


if not CPU_64BITS:
    class user_fpxregs_struct(register_structure):
        _fields_ = (
            ("cwd", c_ushort),
            ("swd", c_ushort),
            ("twd", c_ushort),
            ("fop", c_ushort),
            ("fip", c_ulong),
            ("fcs", c_ulong),
            ("foo", c_ulong),
            ("fos", c_ulong),
            ("mxcsr", c_ulong),
            ("reserved", c_ulong),
            ("st_space", c_ulong * 32),
            ("xmm_space", c_ulong * 32),
            ("padding", c_ulong * 56)
        )

# From /usr/include/asm-generic/siginfo.h


class _sifields_sigfault_t(Union):
    _fields_ = (
        ("_addr", c_void_p),
    )


class _sifields_sigchld_t(Structure):
    _fields_ = (
        ("pid", pid_t),
        ("uid", uid_t),
        ("status", c_int),
        ("utime", clock_t),
        ("stime", clock_t),
    )


class _sifields_t(Union):
    _fields_ = (
        ("pad", c_char * (128 - 3 * sizeof(c_int))),
        ("_sigchld", _sifields_sigchld_t),
        ("_sigfault", _sifields_sigfault_t),
        #        ("_kill", _sifields_kill_t),
        #        ("_timer", _sifields_timer_t),
        #        ("_rt", _sifields_rt_t),
        #        ("_sigpoll", _sifields_sigpoll_t),
    )


class siginfo(Structure):
    _fields_ = (
        ("si_signo", c_int),
        ("si_errno", c_int),
        ("si_code", c_int),
        ("_sifields", _sifields_t)
    )
    _anonymous_ = ("_sifields",)


class iovec_struct(Structure):
    _fields_ = (
        ("buf", c_void_p),
        ("len", c_size_t)
    )



from ptrace.syscall import PtraceSyscall
from signal import SIGTRAP


class SyscallState(object):

    def __init__(self, process):
        self.process = process
        self.ignore_exec_trap = True
        self.ignore_callback = None
        self.clear()

    def event(self, options):
        if self.next_event == "exit":
            return self.exit()
        else:
            return self.enter(options)

    def enter(self, options):
        # syscall enter
        regs = self.process.getregs()
        self.syscall = PtraceSyscall(self.process, options, regs)
        self.name = self.syscall.name
        if (not self.ignore_callback) \
                or (not self.ignore_callback(self.syscall)):
            self.syscall.enter(regs)
        else:
            self.syscall = None
        self.next_event = "exit"
        return self.syscall

    def exit(self):
        if self.syscall:
            self.syscall.exit()
        if self.ignore_exec_trap \
                and self.name == "execve" \
                and not self.process.debugger.trace_exec:
            # Ignore the SIGTRAP after exec() syscall exit
            self.process.syscall()
            self.process.waitSignals(SIGTRAP, SIGTRAP | 0x80)
        syscall = self.syscall
        self.clear()
        return syscall

    def clear(self):
        self.syscall = None
        self.name = None
        self.next_event = "enter"



from ptrace.ctypes_tools import formatAddress, formatWordHex
from ptrace.cpu_info import CPU_WORD_SIZE, CPU_MAX_UINT
from ptrace import PtraceError


class BacktraceFrame(object):
    """
    Backtrace frame.

    Attributes:
     - ip: instruction pointer
     - name: name of the function
     - arguments: value of the arguments
    """

    def __init__(self, ip):
        self.ip = ip
        self.name = u"???"
        self.arguments = []

    def __str__(self):
        arguments = (formatWordHex(arg) for arg in self.arguments)
        return u"IP=%s: %s (%s)" % (formatAddress(self.ip), self.name, ", ".join(arguments))


class Backtrace(object):
    """
    Backtrace: all process frames since the start function.
    """

    def __init__(self):
        self.frames = []
        self.truncated = False

    def append(self, frame):
        self.frames.append(frame)

    def __iter__(self):
        return iter(self.frames)

    def __len__(self):
        return len(self.frames)


def getBacktrace(process, max_args=6, max_depth=20):
    """
    Get the current backtrace of the specified process:
     - max_args: maximum number of arguments in a frame
     - max_depth: maximum number of frames

    Return a Backtrace object.
    """
    backtrace = Backtrace()

    # Get current instruction and frame pointer
    ip = process.getInstrPointer()
    fp = process.getFramePointer()
    depth = 0
    while True:
        # Hit maximum trace depth?
        if max_depth <= depth:
            backtrace.truncated = True
            break

        # Read next frame pointer
        try:
            nextfp = process.readWord(fp)
        except PtraceError:
            nextfp = None

        # Guess number of function argument
        if fp and nextfp:
            nargs = ((nextfp - fp) // CPU_WORD_SIZE) - 2
            nargs = min(nargs, max_args)
        else:
            nargs = 0

        # Create frame
        frame = getBacktraceFrame(process, ip, fp, nargs)
        backtrace.append(frame)

        # End of the stack?
        if not nextfp:
            break

        # Move to next instruction/frame pointer
        ip = process.readWord(fp + CPU_WORD_SIZE)
        if ip == CPU_MAX_UINT:
            # Linux hack to detect end of the stack
            break
        fp = nextfp
        depth += 1
    return backtrace


def getBacktraceFrame(process, ip, fp, nargs):
    """
    Get a backtrace frame:
     - ip: instruction pointer
     - fp: frame pointer
     - nargs: number of arguments

    Return a BacktraceFrame object.
    """
    frame = BacktraceFrame(ip)
    address = fp + CPU_WORD_SIZE
    try:
        for index in range(nargs):
            address += CPU_WORD_SIZE
            word = process.readWord(address)
            frame.arguments.append(word)
    except PtraceError:
        # Ignore argument read error
        pass
    return frame



from ptrace.tools import readBits, formatBits
from ptrace.signames import signalName
from ptrace.ctypes_tools import uint2int

# From /usr/include/bits/mman.h (Ubuntu Feisty, i386)
MMAP_PROT_BITMASK = (
    (1, "PROT_READ"),
    (2, "PROT_WRITE"),
    (4, "PROT_EXEC"),
    (0x01000000, "PROT_GROWSDOWN"),
    (0x02000000, "PROT_GROWSUP"),
)


def formatMmapProt(argument):
    return formatBits(argument.value, MMAP_PROT_BITMASK, "PROT_NONE")


# From /usr/include/bits/mman.h (Ubuntu Feisty, i386)
ACCESS_MODE_BITMASK = (
    (1, "X_OK"),
    (2, "W_OK"),
    (4, "R_OK"),
)


def formatAccessMode(argument):
    return formatBits(argument.value, ACCESS_MODE_BITMASK, "F_OK")


# From /usr/include/bits/fcntl.h (Ubuntu Feisty, i386)
OPEN_FLAGS_BITMASK = [
    (0o1, "O_WRONLY"),
    (0o2, "O_RDWR"),
    (0o100, "O_CREAT"),
    (0o200, "O_EXCL"),
    (0o400, "O_NOCTTY"),
    (0o1000, "O_TRUNC"),
    (0o2000, "O_APPEND"),
    (0o4000, "O_NONBLOCK"),
    (0o10000, "O_SYNC"),
    (0o20000, "O_ASYNC"),
    (0o40000, "O_DIRECT"),
    (0o100000, "O_LARGEFILE"),
    (0o200000, "O_DIRECTORY"),
    (0o400000, "O_NOFOLLOW"),
    (0o1000000, "O_NOATIME"),
    (0o2000000, "O_CLOEXEC"),
    (0o10000000, "O_PATH"),  # Linux 2.6.39
    (0o20200000, "O_TMPFILE"),  # Linux 3.11
]


def formatOpenFlags(argument):
    value = argument.value
    flags = readBits(int(value), OPEN_FLAGS_BITMASK)

    # Add default access mode if neither of the others are present.
    if not flags or flags[0] not in ("O_WRONLY", "O_RDWR"):
        flags.insert(0, "O_RDONLY")

    text = "|".join(flags)
    if value:
        text = "%s (%s)" % (text, oct(argument.value))
    return text


OPEN_MODE_BITMASK = [
    (0o0400, "S_IRUSR"),
    (0o0200, "S_IWUSR"),
    (0o0100, "S_IXUSR"),
    (0o0040, "S_IRGRP"),
    (0o0020, "S_IWGRP"),
    (0o0010, "S_IXGRP"),
    (0o0004, "S_IROTH"),
    (0o0002, "S_IWOTH"),
    (0o0001, "S_IXOTH"),
]


def formatOpenMode(argument):
    value = argument.value
    flags = readBits(int(value), OPEN_MODE_BITMASK)

    text = "|".join(flags)

    # Overwrite text, add it depending on verbosity later
    if value:
        text = "%s" % oct(argument.value)
    return text


CLONE_FLAGS_BITMASK = (
    (0x00000100, "CLONE_VM"),
    (0x00000200, "CLONE_FS"),
    (0x00000400, "CLONE_FILES"),
    (0x00000800, "CLONE_SIGHAND"),
    (0x00002000, "CLONE_PTRACE"),
    (0x00004000, "CLONE_VFORK"),
    (0x00008000, "CLONE_PARENT"),
    (0x00010000, "CLONE_THREAD"),
    (0x00020000, "CLONE_NEWNS"),
    (0x00040000, "CLONE_SYSVSEM"),
    (0x00080000, "CLONE_SETTLS"),
    (0x00100000, "CLONE_PARENT_SETTID"),
    (0x00200000, "CLONE_CHILD_CLEARTID"),
    (0x00400000, "CLONE_DETACHED"),
    (0x00800000, "CLONE_UNTRACED"),
    (0x01000000, "CLONE_CHILD_SETTID"),
    (0x02000000, "CLONE_STOPPED"),
    (0x04000000, "CLONE_NEWUTS"),
    (0x08000000, "CLONE_NEWIPC"),
)


def formatCloneFlags(argument):
    flags = argument.value
    bits = readBits(flags, CLONE_FLAGS_BITMASK)
    signum = flags & 0xFF
    if signum:
        bits.insert(0, signalName(signum))
    if bits:
        bits = "%s" % ("|".join(bits))
        return "<%s> (%s)" % (bits, str(flags))
    else:
        return str(flags)


AT_FDCWD = -100


def formatDirFd(argument):
    value = uint2int(argument.value)
    return "AT_FDCWD" if value == AT_FDCWD else str(value)



#!/usr/bin/python
from ptrace import PtraceError
from ptrace.debugger import (PtraceDebugger, Application,
                             ProcessExit, NewProcessEvent, ProcessSignal,
                             ProcessExecution, ProcessError)
from optparse import OptionParser
from os import getpid
from sys import stdout, stderr, exit
from logging import getLogger, info, warning, error
from ptrace.version import __version__ as VERSION, WEBSITE
from ptrace.error import PTRACE_ERRORS, writeError
from ptrace.binding import HAS_PTRACE_SINGLESTEP
from ptrace.disasm import HAS_DISASSEMBLER
from ptrace.ctypes_tools import (truncateWord,
                                 formatWordHex, formatAddress, formatAddressRange, word2bytes)
from ptrace.process_tools import dumpProcessInfo
from ptrace.tools import inverseDict
from ptrace.func_call import FunctionCallOptions
from ptrace.signames import signalName, SIGNAMES
from signal import SIGTRAP, SIGINT
from ptrace.terminal import enableEchoMode, terminalWidth
from errno import ESRCH
from ptrace.cpu_info import CPU_POWERPC
from ptrace.debugger import ChildError
from ptrace.debugger.memory_mapping import readProcessMappings
from ptrace.os_tools import RUNNING_PYTHON3
try:
    unichr
    raw_input
except NameError:
    # Python 3
    unichr = chr
    raw_input = input

import re
if stdout.isatty():
    try:
        # Use readline for better raw_input()
        import readline   # noqa
    except ImportError:
        pass

# Match a register name: $eax, $gp0, $orig_eax
REGISTER_REGEX = re.compile(r"\$[a-z]+[a-z0-9_]+")

# BYTES_REGEX = re.compile(r"""(?:'([^'\\]*)'|"([^"\\]*)")""")

SIGNALS = inverseDict(SIGNAMES)   # name -> signum

COMMANDS = (
    # trace instructions
    ("cont", "continue execution"),
    ("step", "execute one instruction (do not enter in a call)"),
    ("stepi", "execute one instruction (enter the call)"),
    ("until", "execute code until specified address (until <address>)"),
    ("set", "set register value (set <register>=<value>)"),
    ("sys", "continue execution to next syscall"),
    ("signal", "send a signal to the process (signal <signum>)"),
    ("signals", "display signals"),

    # current process info
    ("regs", "display registers"),
    ("where", "display true code content (show breakpoints effects on code). e.g. 'where $eip', 'where $eip $eip+20'"),
    ("print", "display a value (print <value>)"),
    ("hexdump", "dump memory as specified address or address range (hexdump <address> or hexdump <start> <stop>)"),
    ("gcore", "generate core file of the process"),
    ("where2", "display original code content (don't show effects of breakpoint on code)"),
    ("stack", "display stack content"),
    ("backtrace", "dump the backtrace"),
    ("proc", "display process information"),
    ("maps", "display memory mappings"),

    # breakpoints
    ("break", "set a breakpoint (break <address>)"),
    ("breakpoints", "display breakpoints"),
    ("delete", "delete a breakpoint (delete <address>)"),

    # processes
    ("attach", 'attach a new process (e.g. "attach 2390")'),
    ("proclist", "list of traced processes"),
    ("switch", "switch active process (switch or switch <pid>)"),

    ("follow", r'''follow a term (e.g. "follow '\x12\x14\x27\x13'")'''),
    ("showfollow", 'show all "followed" terms'),
    ("resetfollow", 'reset all "followed" terms'),
    ("xray", 'show addresses of (and possible pointers to) "followed" terms'),

    # other
    ("dbginfo", "information about the debugger"),
    ("quit", "quit debugger"),
    ("help", "display this help"),
)


def formatAscii(data):
    def asciiChar(byte):
        if 32 <= byte <= 126:
            return unichr(byte)
        else:
            return '.'
    if RUNNING_PYTHON3:
        return u''.join(asciiChar(byte) for byte in data)
    else:
        return u''.join(asciiChar(ord(byte)) for byte in data)


def formatHexa(data):
    if RUNNING_PYTHON3:
        return u' '.join(u"%02x" % byte for byte in data)
    else:
        return u' '.join(u"%02x" % ord(byte) for byte in data)

# finds possible pointer values in process memory space,
# pointing to address


def getPointers(process, address):
    address = word2bytes(address)
    procmaps = readProcessMappings(process)
    for pm in procmaps:
        for found in pm.search(address):
            yield found


class Gdb(Application):

    def __init__(self):
        Application.__init__(self)

        # Parse self.options
        self.parseOptions()

        # Setup output (log)
        self.setupLog()

        self.last_signal = {}

        # We assume user wants all possible information
        self.syscall_options = FunctionCallOptions(
            write_types=True,
            write_argname=True,
            write_address=True,
        )

        # FIXME: Remove self.breaks!
        self.breaks = dict()

        self.followterms = []

    def setupLog(self):
        self._setupLog(stdout)

    def parseOptions(self):
        parser = OptionParser(
            usage="%prog [options] -- program [arg1 arg2 ...]")
        self.createCommonOptions(parser)
        self.createLogOptions(parser)
        self.options, self.program = parser.parse_args()

        if self.options.pid is None and not self.program:
            parser.print_help()
            exit(1)

        self.processOptions()
        self.show_pid = self.options.fork

    def _continueProcess(self, process, signum=None):
        if not signum and process in self.last_signal:
            signum = self.last_signal[process]

        if signum:
            error("Send %s to %s" % (signalName(signum), process))
            process.cont(signum)
            try:
                del self.last_signal[process]
            except KeyError:
                pass
        else:
            process.cont()

    def cont(self, signum=None):
        for process in self.debugger:
            process.syscall_state.clear()
            if process == self.process:
                self._continueProcess(process, signum)
            else:
                self._continueProcess(process)

        # Wait for a process signal
        signal = self.debugger.waitSignals()
        process = signal.process

        # Hit breakpoint?
        if signal.signum == SIGTRAP:
            ip = self.process.getInstrPointer()
            if not CPU_POWERPC:
                # Go before "INT 3" instruction
                ip -= 1
            breakpoint = self.process.findBreakpoint(ip)
            if breakpoint:
                error("Stopped at %s" % breakpoint)
                breakpoint.desinstall(set_ip=True)
        else:
            self.processSignal(signal)
        return None

    def readRegister(self, regs):
        name = regs.group(0)[1:]
        value = self.process.getreg(name)
        return str(value)

    def parseInteger(self, text):
        # Remove spaces and convert to lower case
        text = text.strip()
        if " " in text:
            raise ValueError("Space are forbidden: %r" % text)
        text = text.lower()

        # Replace registers by their value
        orig_text = text
        text = REGISTER_REGEX.sub(self.readRegister, text)

        # Replace hexadecimal numbers by decimal numbers
        def readHexadecimal(regs):
            text = regs.group(0)
            if text.startswith("0x"):
                text = text[2:]
            elif not re.search("[a-f]", text):
                return text
            value = int(text, 16)
            return str(value)
        text = re.sub(r"(?:0x)?[0-9a-f]+", readHexadecimal, text)

        # Reject invalid characters
        if not re.match(r"^[()<>+*/&0-9-]+$", text):
            raise ValueError("Invalid expression: %r" % orig_text)

        # Use integer division (a//b) instead of float division (a/b)
        text = text.replace("/", "//")

        # Finally, evaluate the expression
        is_pointer = text.startswith("*")
        if is_pointer:
            text = text[1:]
        try:
            value = eval(text)
            value = truncateWord(value)
        except SyntaxError:
            raise ValueError("Invalid expression: %r" % orig_text)
        if is_pointer:
            value = self.process.readWord(value)
        return value

    def parseIntegers(self, text):
        values = []
        for item in text.split():
            item = item.strip()
            value = self.parseInteger(item)
            values.append(value)
        return values

    def parseBytes(self, text):
        # FIXME: Validate input
        #        if not BYTES_REGEX.match(text):
        #            raise ValueError('Follow text must be enclosed in quotes!')
        text = 'b' + text.lstrip()
        value = eval(text)
        if not isinstance(value, bytes):
            raise TypeError("Input is not a bytes string!")
        return value

    def addFollowTerm(self, text):
        # Allow terms of the form 'string', "string", '\x04', "\x01\x14"
        term = self.parseBytes(text)
        self.followterms.append(term)

    def showFollowTerms(self):
        print(self.followterms)

    def _xray(self):
        for term in self.followterms:
            for process in self.debugger:
                for procmap in readProcessMappings(process):
                    for address in procmap.search(term):
                        yield (process, procmap, address, term)

    # displays the offsets of all terms found in the process memory mappings
    # along with possible addresses of pointers pointing to these terms
    def xray(self):
        for process, procmap, address, term in self._xray():
            pointers = " ".join(formatAddress(ptr_addr)
                                for ptr_addr in getPointers(process, address))
            print("term[%s] pid[%i] %s %s pointers: %s" % (
                repr(term), process.pid, procmap,
                formatAddress(address),
                pointers))

    def execute(self, command):
        errmsg = None
        if command == "cont":
            errmsg = self.cont()
        elif command == "proc":
            self.procInfo()
        elif command == "proclist":
            self.procList()
        elif command.startswith("attach "):
            errmsg = self.attachProcess(command[7:])
        elif command == "regs":
            self.process.dumpRegs()
        elif command == "stack":
            self.process.dumpStack()
        elif command == "gcore":
            self.gcore(self.process)
        elif command == "backtrace":
            errmsg = self.backtrace()
        elif command == "where" or command.startswith("where "):
            errmsg = self.where(command[6:])
        elif command == "where2" or command.startswith("where2 "):
            errmsg = self.where(command[7:], manage_bp=True)
        elif command == "maps":
            self.process.dumpMaps()
        elif command == "dbginfo":
            self.debuggerInfo()
        elif command == "step":
            errmsg = self.step(False)
        elif command == "stepi":
            errmsg = self.step(True)
        elif command == "sys":
            errmsg = self.syscallTrace()
        elif command == "help":
            self.help()
        elif command.startswith("set "):
            errmsg = self.set(command)
        elif command.startswith("until "):
            errmsg = self.until(command[6:])
        elif command.startswith("switch") or command == "switch":
            errmsg = self.switch(command[6:])
        elif command.startswith("break "):
            errmsg = self.breakpoint(command[6:])
        elif command.startswith("breakpoints"):
            self.displayBreakpoints()
        elif command.startswith("signals"):
            self.displaySignals()
        elif command.startswith("delete "):
            errmsg = self.delete(command[7:])
        elif command.startswith("hexdump "):
            errmsg = self.hexdump(command[8:])
        elif command.startswith("signal "):
            errmsg = self.signal(command[7:])
        elif command.startswith("print "):
            errmsg = self.print_(command[6:])
        elif command.startswith("follow "):
            errmsg = self.addFollowTerm(command[7:])
        elif command == "showfollow":
            self.showFollowTerms()
        elif command == "resetfollow":
            self.followterms = []
        elif command == "xray":
            self.xray()
        else:
            errmsg = "Unknown command: %r" % command
        if errmsg:
            print(errmsg, file=stderr)
            return False
        return True

    def parseSignum(self, command):
        try:
            return SIGNALS[command]
        except KeyError:
            pass
        try:
            return SIGNALS["SIG" + command]
        except KeyError:
            pass
        try:
            return self.parseInteger(command)
        except ValueError:
            raise ValueError("Invalid signal number: %r" % command)

    def signal(self, command):
        try:
            signum = self.parseSignum(command)
        except ValueError as err:
            return str(err)
        last_process = self.process
        try:
            errmsg = self.cont(signum)
            return errmsg
        finally:
            try:
                del self.last_signal[last_process]
            except KeyError:
                pass

    def print_(self, command):
        try:
            value = self.parseInteger(command)
        except ValueError as err:
            return str(err)
        error("Decimal: %s" % value)
        error("Hexadecimal: %s" % formatWordHex(value))
        for map in self.process.readMappings():
            if value not in map:
                continue
            error("Address is part of mapping: %s" % map)
        return None

    def gcore(self, process):
        import re
        childPid = str(process).split('#')[-1][:-1]
        maps_file = open("/proc/" + childPid + "/maps", 'r')
        mem_file = open("/proc/" + childPid + "/mem", 'r', 0)
        from sys import argv
        dump = open("/vmdump/" + argv[1] + ".dump", "wb")
        for line in maps_file.readlines():  # for each mapped region
            m = re.match(r'([0-9A-Fa-f]+)-([0-9A-Fa-f]+) ([-r])', line)
            if m.group(3) == 'r':  # if this is a readable region
                if "/lib" not in line and "/usr" not in line:  # for eliminating of shared libs
                    start = int(m.group(1), 16)
                    end = int(m.group(2), 16)
                    mem_file.seek(start)  # seek to region start
                    chunk = mem_file.read(end - start)  # read region contents
                    dump.write(chunk,)  # dump contents to standard output
        maps_file.close()
        mem_file.close()
        dump.close()

    def hexdump(self, command):
        max_line = 20
        width = (terminalWidth() - len(formatAddress(1)) - 3) // 4
        width = max(width, 1)

        limited = None
        parts = command.split(" ", 1)
        if 1 < len(parts):
            try:
                start_address = self.parseInteger(parts[0])
                end_address = self.parseInteger(parts[1])
                if end_address <= start_address:
                    raise ValueError('End address (%s) is smaller than start address(%s)!'
                                     % (formatAddress(end_address), formatAddress(start_address)))
            except ValueError as err:
                return str(err)
            size = end_address - start_address
            max_size = width * max_line
            if max_size < size:
                limited = max_size
                end_address = start_address + max_size
        else:
            try:
                start_address = self.parseInteger(command)
            except ValueError as err:
                return str(err)
            end_address = start_address + 5 * width

        read_error = None
        address = start_address
        while address < end_address:
            size = min(end_address - address, width)
            try:
                # Read bytes
                memory = self.process.readBytes(address, size)

                # Format bytes
                hexa = formatHexa(memory)
                hexa = hexa.ljust(width * 3 - 1, ' ')

                ascii = formatAscii(memory)
                ascii = ascii.ljust(width, ' ')

                # Display previous read error, if any
                if read_error:
                    warning("Warning: Unable to read memory %s" % (
                        formatAddressRange(*read_error)))
                    read_error = None

                # Display line
                error("%s| %s| %s" % (formatAddress(address), hexa, ascii))
            except PtraceError:
                if not read_error:
                    read_error = [address, address + size]
                else:
                    read_error[1] = address + size
            address += size

        # Display last read error, if any
        if read_error:
            warning("Warning: Unable to read memory %s" % (
                formatAddressRange(*read_error)))
        if limited:
            warning("(limit to %s bytes)" % max_size)
        return None

    def backtrace(self):
        trace = self.process.getBacktrace()
        for func in trace:
            error(func)
        if trace.truncated:
            error("--limited to depth %s--" % len(trace))
        return None

    def where(self, command, manage_bp=False):
        start = None
        stop = None
        try:
            values = self.parseIntegers(command)
        except ValueError as err:
            return str(err)
        if 1 <= len(values):
            start = values[0]
        if 2 <= len(values):
            stop = values[1]
        self.process.dumpCode(start, stop, manage_bp=manage_bp)
        return None

    def procInfo(self):
        dumpProcessInfo(error, self.process.pid, max_length=160)

    def procList(self):
        for process in self.debugger:
            text = str(process)
            if self.process == process:
                text += " (active)"
            error(text)

    def set(self, command):
        try:
            key, value = command[4:].split("=", 1)
            key = key.strip().lower()
            if not key.startswith("$"):
                return 'Register name (%s) have to start with "$"' % key
            key = key[1:]
        except ValueError:
            return "Invalid command: %r" % command
        try:
            value = self.parseInteger(value)
        except ValueError as err:
            return str(err)
        try:
            self.process.setreg(key, value)
        except ProcessError as err:
            return "Unable to set $%s=%s: %s" % (key, value, err)
        error("Set $%s to %s" % (key, value))
        return None

    def displayInstr(self, prefix):
        try:
            if HAS_DISASSEMBLER:
                instr = self.process.disassembleOne()
                error("%s %s: %s" % (
                    prefix, formatAddress(instr.address), instr.text))
            else:
                self.process.dumpCode()
        except PtraceError as err:
            error("Unable to read current instruction: %s" % err)

    def attachProcess(self, text):
        try:
            pid = self.parseInteger(text)
        except ValueError as err:
            return str(err)
        process = self.debugger.addProcess(pid, False)
        self.switchProcess(process)

    def step(self, enter_call, address=None):
        if address is None:
            self.displayInstr("Execute")
        if (not HAS_PTRACE_SINGLESTEP) or (not enter_call):
            if address is None:
                address = self.process.getInstrPointer()
                size = self.readInstrSize(address, default_size=None)
                if not size:
                    return "Unable to read instruction size at %s" \
                        % formatAddress(address)
                address += size
            size = self.readInstrSize(address)

            # Set a breakpoint
            breakpoint = self.process.createBreakpoint(address, size)

            # Continue the process
            self.process.cont()
        else:
            # Use ptrace single step command
            self.process.singleStep()
            breakpoint = None

        # Execute processus until next TRAP
        try:
            self.process.waitSignals(SIGTRAP)
            if breakpoint:
                breakpoint.desinstall(set_ip=True)
        except:   # noqa: E722
            if breakpoint:
                breakpoint.desinstall()
            raise
        return None

    def newProcess(self, event):
        error("New process: %s" % event.process)

    # FIXME: This function doesn't work with multiple processes
    # especially when a parent waits for a child
    def syscallTrace(self):
        # Trace until syscall enter
        self.process.syscall()
        self.process.waitSyscall()

        # Process the syscall event
        state = self.process.syscall_state
        syscall = state.event(self.syscall_options)

        # Display syscall
        if syscall:
            if syscall.result is not None:
                text = "%s = %s" % (syscall.format(), syscall.result_text)
                if self.show_pid:
                    text = "Process %s exits %s" % (syscall.process.pid, text)
                error(text)
            else:
                text = syscall.format()
                if self.show_pid:
                    text = "Process %s enters %s" % (syscall.process.pid, text)
                error(text)
        return None

    def until(self, command):
        try:
            address = self.parseInteger(command)
        except ValueError as err:
            return str(err)
        errmsg = self.step(False, address)
        if errmsg:
            return errmsg
        self.displayInstr("Current")
        return None

    def switch(self, command):
        if not command:
            process_list = self.debugger.list
            if len(process_list) == 1:
                return "There is only one process!"
            index = process_list.index(self.process)
            index = (index + 1) % len(process_list)
            process = process_list[index]
            self.switchProcess(process)
            return
        try:
            pid = self.parseInteger(command)
        except ValueError as err:
            return str(err)
        try:
            process = self.debugger[pid]
            self.switchProcess(process)
        except KeyError:
            return "There is not process %s" % pid
        return None

    def switchProcess(self, process):
        if self.process == process:
            return
        error("Switch to %s" % process)
        self.process = process

    def nextProcess(self):
        try:
            process = next(iter(self.debugger))
            self.switchProcess(process)
        except StopIteration:
            pass

    def displayBreakpoints(self):
        found = False
        for process in self.debugger:
            for bp in process.breakpoints.values():
                found = True
                error("%s:%s" % (process, bp))
        if not found:
            error("(no breakpoint)")

    def displaySignals(self):
        signals = list(SIGNAMES.items())
        signals.sort(key=lambda key_value: key_value[0])
        for signum, name in signals:
            error("% 2s: %s" % (signum, name))

    def readInstrSize(self, address, default_size=None):
        if not HAS_DISASSEMBLER:
            return default_size
        try:
            # Get address and size of instruction at specified address
            instr = self.process.disassembleOne(address)
            return instr.size
        except PtraceError as err:
            warning("Warning: Unable to read instruction size at %s: %s" % (
                formatAddress(address), err))
            return default_size

    def breakpoint(self, command):
        try:
            address = self.parseInteger(command)
        except ValueError as err:
            return str(err)

        # Create breakpoint
        size = self.readInstrSize(address)
        try:
            bp = self.process.createBreakpoint(address, size)
        except PtraceError as err:
            return "Unable to set breakpoint at %s: %s" % (
                formatAddress(address), err)
        error("New breakpoint: %s" % bp)
        return None

    def delete(self, command):
        try:
            address = self.parseInteger(command)
        except ValueError as err:
            return str(err)

        breakpoint = self.process.findBreakpoint(address)
        if not breakpoint:
            return "No breakpoint at %s " % formatAddress(address)
        breakpoint.desinstall()
        error("%s deleted" % breakpoint)
        return None

    def help(self):
        for command, description in COMMANDS:
            error("%s: %s" % (command, description))
        error('')
        error("Value can be an hexadecimal/decimal number or a register name ($reg)")
        error("You can use operators a+b, a-b, a*b, a/b, a<<b, a>>b, a**b, and parenthesis in expressions")
        error(
            'Use ";" to write multiple commands on the same line (e.g. "step; print $eax")')

    def processSignal(self, event):
        event.display()
        self.switchProcess(event.process)
        self.last_signal[self.process] = event.signum
        error("%s interrupted by %s" % (self.process, event.name))

    def processExecution(self, event):
        error(event)
        self.switchProcess(event.process)
        self.interrupt()

    def debuggerInfo(self):
        error("Debugger process ID: %s" % getpid())
        error("python-ptrace version %s" % VERSION)
        error("Website: %s" % WEBSITE)

    def interrupt(self):
        waitlist = []
        for process in self.debugger:
            if process.is_stopped:
                continue
            try:
                if process.isTraced():
                    continue
            except NotImplementedError:
                pass
            warning("Interrupt %s (send SIGINT)" % process)
            process.kill(SIGINT)
            waitlist.append(process)
        for process in waitlist:
            info("Wait %s interruption" % process)
            try:
                process.waitSignals(SIGINT)
            except ProcessSignal as event:
                event.display()
            except KeyboardInterrupt:
                pass

    def deleteProcess(self, pid):
        try:
            process = self.debugger[pid]
        except KeyError:
            return
        event = process.processTerminated()
        error(str(event))
        if process == self.process:
            self.nextProcess()

    def restoreTerminal(self):
        if enableEchoMode():
            error("Terminal: restore echo mode")

    def mainLoop(self):
        # Read command
        try:
            self.restoreTerminal()
            command = raw_input(self.invite).strip()
        except EOFError:
            print()
            return True
        except KeyboardInterrupt:
            error("User interrupt!")
            self.interrupt()
            return False

        # If command is empty, reuse previous command
        if not command:
            if self.previous_command:
                command = self.previous_command
                info("Replay previous command: %s" % command)
            else:
                return False
        self.previous_command = None

        # User wants to quit?
        if command == "quit":
            return True

        # Execute the user command
        try:
            command_str = command
            ok = True
            for command in command_str.split(";"):
                command = command.strip()
                try:
                    ok &= self.execute(command)
                except Exception as err:
                    print("Command error: %s" % err)
                    raise
                    ok = False
                if not ok:
                    break
            if ok:
                self.previous_command = command_str
        except KeyboardInterrupt:
            self.interrupt()
        except NewProcessEvent as event:
            self.newProcess(event)
        except ProcessSignal as event:
            self.processSignal(event)
        except ProcessExit as event:
            error(event)
            self.nextProcess()
        except ProcessExecution as event:
            self.processExecution(event)
        except PtraceError as err:
            error("ERROR: %s" % err)
            if err.errno == ESRCH:
                self.deleteProcess(err.pid)
        return False

    def runDebugger(self):
        self.setupDebugger()

        # Create new process
        try:
            self.process = self.createProcess()
        except ChildError as err:
            writeError(getLogger(), err, "Unable to create child process")
            return
        if not self.process:
            return

        # Trace syscalls
        self.invite = '(gdb) '
        self.previous_command = None
        while True:
            if not self.debugger:
                # There is no more process: quit
                return
            done = self.mainLoop()
            if done:
                return

    def main(self):
        self.debugger = PtraceDebugger()
        try:
            self.runDebugger()
        except KeyboardInterrupt:
            error("Interrupt debugger: quit!")
        except PTRACE_ERRORS as err:
            writeError(getLogger(), err, "Debugger error")
        self.process = None
        self.debugger.quit()
        error("Quit gdb.")
        self.restoreTerminal()


if __name__ == "__main__":
    Gdb().main()



from ptrace.cpu_info import CPU_WORD_SIZE
from ptrace.ctypes_tools import ntoh_ushort, ntoh_uint  # noqa
from ptrace.syscall import SYSCALL_PROTOTYPES
from ptrace.syscall.socketcall_constants import SOCKETCALL, SOCKET_FAMILY  # noqa
from ptrace.syscall.socketcall_struct import sockaddr, sockaddr_in, sockaddr_in6, sockaddr_un
from ctypes import c_int
from ptrace.os_tools import RUNNING_LINUX
from socket import AF_INET, AF_INET6, inet_ntoa   # noqa
if RUNNING_LINUX:
    from socket import AF_NETLINK
    from ptrace.syscall.socketcall_struct import sockaddr_nl

AF_FILE = 1


def formatOptVal(argument):
    function = argument.function
    optlen = function["optlen"].value
    if optlen == 4:
        addr = argument.value
        text = function.process.readStruct(addr, c_int)
        return argument.formatPointer("<%s>" % text, addr)
    else:
        return None


def formatSockaddr(argument, argtype):
    address = argument.value
    value = argument.function.process.readStruct(address, sockaddr)
    family = value.family
    if family == AF_INET:
        return argument.readStruct(address, sockaddr_in)
    if family == AF_INET6:
        return argument.readStruct(address, sockaddr_in6)
    if family == AF_FILE:
        return argument.readStruct(address, sockaddr_un)
    if RUNNING_LINUX:
        if family == AF_NETLINK:
            return argument.readStruct(address, sockaddr_nl)
    family = SOCKET_FAMILY.get(family, family)
    return argument.formatPointer("<sockaddr family=%s>" % family, address)


def setupSocketCall(function, process, socketcall, address):
    # Reset function call
    function.clearArguments()
#    function.argument_class = SocketCallArgument

    # Setup new function call
    function.process = process
    function.name = socketcall.getText()

    # Create arguments
    function.restype, formats = SYSCALL_PROTOTYPES[function.name]
    for argtype, argname in formats:
        value = process.readWord(address)
        function.addArgument(value, argname, argtype)
        address += CPU_WORD_SIZE


def formatSockaddrInStruct(argument, name, value):
    if name == "sin_port":
        return ntoh_ushort(value)
    return None


def formatSockaddrIn6Struct(argument, name, value):
    if name == "sin6_port":
        return ntoh_ushort(value)
    # if name == "sin6_addr":
        # FIXME: ...
    return None



"""
Terminal functions.
"""

from termios import tcgetattr, tcsetattr, ECHO, TCSADRAIN, TIOCGWINSZ
from sys import stdin, stdout
from fcntl import ioctl
from struct import unpack
import os

TERMIO_LFLAGS = 3


def _terminalSize():
    fd = stdout.fileno()
    size = ioctl(fd, TIOCGWINSZ, '1234')
    height, width = unpack('hh', size)
    return (width, height)


def terminalWidth():
    """
    Get the terminal width in characters.
    """
    return _terminalSize()[0]


def enableEchoMode():
    """
    Enable echo mode in the terminal. Return True if the echo mode is set
    correctly, or False if the mode was already set.
    """
    fd = stdin.fileno()
    if not os.isatty(fd):
        return False
    state = tcgetattr(fd)
    if state[TERMIO_LFLAGS] & ECHO:
        return False
    state[TERMIO_LFLAGS] = state[TERMIO_LFLAGS] | ECHO
    tcsetattr(fd, TCSADRAIN, state)
    return True
